/*

*/

import java.util.concurrent.locks.ReentrantLock;

/**
   This simulates a first-in-first-out (stack) buffer.
*/
public class Buffer_locked extends Buffer
{
   private final ReentrantLock lock;

   /**
      Create a buffer with a mutual exclusion lock.

      @param size          the size of the buffer
      @param currentIndex  initially fill the buffer up to this index
      @throws IllegalArgumentException if currentIndex is outside of the buffer
   */
   Buffer_locked(final int size, final int currentIndex)
   {
      super(size, currentIndex); // Initialize a partially filled buffer.

      this.lock = new ReentrantLock();
   }


   /**
      Insert the given character at the next available buffer location.

      Every time we do an operation on this buffer, it prints out its state.

      @param c  character to put in the buffer
      @thows IllegalStateException when the buffer is full
   */
   @Override
   public void put(char c)
   {
      if (currentIndex == size)
      {
         // This buffer is not bounded so this buffer
         // throws an exception when it tries to overflow.
         throw new IllegalStateException("buffer is full");
      }

      this.lock.lock(); // mutual exclusion
      try
      {
         ++currentIndex;
         Thread.yield(); // Allow race conditions to happen more often.
         buf[currentIndex] = c;
         System.out.println(this);  // Print the state of this buffer.
         System.out.flush();
      }
      finally
      {
         this.lock.unlock();
      }
   }


   /**
      Remove from the buffer the character at the buffer's current index.
      Mark the newly emptied spot in the buffer with the given character.
      Return the character that was removed from the buffer.

      Every time we do an operation on this buffer, it prints out its state.

      @param c  character to use to mark the newly emptied position in the buffer
      @return the character removed from the buffer
      @thows IllegalStateException when the buffer is empty
   */
   @Override
   public char get(char c)
   {
      if (currentIndex == 0)
      {
         // This buffer is not bounded so this buffer
         // throws an exception when it tries to underflow.
         throw new IllegalStateException("buffer is empty");
      }

      this.lock.lock(); // mutual exclusion
      try
      {
         char result = buf[currentIndex];
         buf[currentIndex] = c;
         Thread.yield(); // Allow race conditions to happen more often.
         --currentIndex;
         System.out.println(this);  // Print the state of this buffer.
         System.out.flush();
         return result;
      }
      finally
      {
         this.lock.unlock();
      }
   }
}
