/*

*/

import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Condition;

/**
   This simulates a first-in-first-out (stack) buffer.
*/
public class Buffer_locked_bounded extends Buffer
{
   private final ReentrantLock lock;
   private final Condition emptyCondition;
   private final Condition fullCondition;

   /**
      Create a buffer with a mutual exclusion lock,
      a condition variable that prevents buffer
      underflow, and another condition variable
      that prevents buffer overflow.

      @param size          the size of the buffer
      @param currentIndex  initially fill the buffer up to this index
      @throws IllegalArgumentException if currentIndex is outside of the buffer
 */
   Buffer_locked_bounded(final int size, final int currentIndex)
   {
      super(size, currentIndex); // Initialize a partially filled buffer.

      this.lock = new ReentrantLock();
      this.fullCondition  = lock.newCondition();
      this.emptyCondition = lock.newCondition();
   }


   /**
      Insert the given character at the next available buffer location.

      Every time we do an operation on this buffer, it prints out its state.

      @param c  character to put in the buffer
   */
   @Override
   public void put(char c)
   throws InterruptedException
   {
      this.lock.lock(); // mutual exclusion
      try
      {
         while ( this.isFull() )
         {
            this.fullCondition.await();
         }
         ++currentIndex;
         Thread.yield(); // Allow race conditions to happen more often.
         buf[currentIndex] = c;
         this.emptyCondition.signal(); // Can use signalAll() OR signal().
         System.out.println(this);     // Print the state of this buffer.
         System.out.flush();
      }
      finally
      {
         this.lock.unlock();
      }
   }


   /**
      Remove from the buffer the character at the buffer's current index.
      Mark the newly emptied spot in the buffer with the given character.
      Return the character that was removed from the buffer.

      Every time we do an operation on this buffer, it prints out its state.

      @param c  character to use to mark the newly emptied position in the buffer
      @return the character removed from the buffer
   */
   @Override
   public char get(char c)
   throws InterruptedException
   {
      this.lock.lock(); // mutual exclusion
      try
      {
         while ( this.isEmpty() )
         {
            this.emptyCondition.await();
         }
         char result = buf[currentIndex];
         buf[currentIndex] = c;
         Thread.yield(); // Allow race conditions to happen more often.
         --currentIndex;
         this.fullCondition.signal(); // Can use signalAll() OR signal().
         System.out.println(this);    // Print the state of this buffer.
         System.out.flush();
         return result;
      }
      finally
      {
         this.lock.unlock();
      }
   }
}
